# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later

class ConfigurationError(RuntimeError):
    """Indicate an invalid configuration file.

    Usually fatal.
    """


class PaasError(RuntimeError):
    """Used for errors encountered on PAAS."""
    def __init__(self, executor, action, code,
                 transport_code=None,  # used when different from code
                 action_details=None,
                 error_details=None):
        self.executor = executor
        self.action = action
        self.action_details = action_details
        self.transport_code = transport_code
        self.code = code
        self.error_details = error_details
        self.args = (executor, action, code)


class PaasProvisioningError(PaasError):
    """Used for errors encountered while provisioning on PAAS."""


class PaasResourceError(PaasError):
    """Errors encountered on an already provisioned PAAS resource."""

    def __init__(self, resource_id, *args, **kwargs):
        super(PaasResourceError, self).__init__(*args, **kwargs)
        self.resource_id = resource_id
        self.args = (resource_id, ) + self.args


class GitLabError(RuntimeError):
    """Used for errors encountered in interactions with GitLab."""

    def __init__(self, url, status_code, params, message, headers=None):
        self.url = url
        self.status_code = status_code
        self.params = params
        self.message = message
        self.headers = headers
        self.args = (url, status_code, params, message)


class GitLabUnavailableError(GitLabError):
    """Used when API is not reachable.

    This condition is often considered to be non permanent.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('params', None)
        kwargs.setdefault('status_code', None)
        super(GitLabUnavailableError, self).__init__(*args, **kwargs)


class GitLabUnexpectedError(GitLabError):
    """Used for HTTP status code that represent hard errors.
    """


class JobLaunchTimeout(RuntimeError):
    """Used to express that the job wasn't fully started in time.

    By fully started, we mean that the inner executor reported back to
    the coordinator (current way of checking is by accessing the job trace).

    :param *args: (job_handle, timeout)
    """


class DeploymentBranchAlreadyExisting(RuntimeError):
    """Raised when a deployment repository already has a branch.

    It is a common pattern with PAAS providers to trigger deployments
    by Git pushes to a given branch. In the context of Heptapod PAAS Runner,
    applications are oneshots (CI jobs) hence expect to be the very first
    deployments. An existing branch is therefore usually the symptom of a
    problem in state tracking.

    The point is to have a precise, catchable error instead of a generic
    CalledProcessError or similar.
    """



# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later
import argparse
import logging
import os
import requests
import toml

AVAILABLE_EXECUTORS = {'clever-docker': "Clever Cloud Docker",
                       'local-docker': "Local Docker, "
                       "for testing purposes only"}

logger = logging.getLogger(__name__)


def register(conf_path, runner_name, coordinator_url, token, executor):
    """Use token to register runner with given executor and write to conf_path.
    """
    resp = requests.post(coordinator_url.rstrip('/') + '/api/v4/runners',
                         json=dict(info=dict(name=runner_name),
                                   description=runner_name,
                                   active=False,
                                   token=token,
                                   ))

    if resp.status_code >= 400:
        raise RuntimeError(resp.status_code, resp.text)

    runner_token = resp.json()['token']

    runner = dict(
        name=runner_name,
        url=coordinator_url,
        token=runner_token,
        executor=executor)

    with open(conf_path, 'a') as conf_file:
        conf_file.write('\n')
        conf_file.write(toml.dumps(dict(runners=[runner])))


def positive_int(s):
    i = int(s)
    if i <= 0:
        raise TypeError("Not a positive integer: %r" % s)
    return i


def non_empty_string(s):
    if not s or not isinstance(s, str):
        raise TypeError("Empty string (or not a string) "
                        "is not allowed: %r" % s)
    return s


def create_conf_file(path, state_file_path, concurrency):
    with open(path, 'w') as conf_file:
        toml.dump(dict(state_file=state_file_path,
                       concurrent=concurrency),
                  conf_file)


def valid_input(value, prompt="", initial_message=None,
                converter=non_empty_string):
    """Repeatedly ask user for input and convert it until valid.

    :param value: can be ``None``, which is expected to always be invalid
       (argparse default), without user feedback about it. In other invalid
       cases (actual value provided or wrong input), a warning log is issued.
    """
    converted = None
    first = True

    if prompt:
        prompt = prompt.strip() + ': '

    while converted is None:
        try:
            converted = converter(value)
        except (TypeError, ValueError) as exc:
            if value is not None:
                logger.warning(str(exc))

            if first and initial_message:
                print(initial_message)
                first = False

            value = input(prompt).strip()

    return converted


def valid_executor(exe):
    if exe not in AVAILABLE_EXECUTORS:
        raise ValueError("Not an available executor: %r" % exe)
    return exe


def main(raw_args=None):
    """Console script entry point.
    """
    parser = argparse.ArgumentParser(
        description="Initiate a Runner configuration by performing the "
        "GitLab registration",
        epilog="The configuration file will be created if needed "
        "and the new Runner configuration appended to it."
    )
    parser.add_argument("config", help="Path to Heptapod Runner "
                        "configuration file to write into")
    parser.add_argument("--name", help="Runner name")
    parser.add_argument('-c', "--coordinator-url",
                        help="Base URL of the GitLab / Heptapod instance to "
                        "run for")
    parser.add_argument('-t', "--registration-token",
                        help="Registration token, as found in Application "
                        "Group or Project settings")
    parser.add_argument('-e', "--executor",
                        choices=AVAILABLE_EXECUTORS.keys())
    init_conf_grp = parser.add_argument_group(
        title="Configuration creation options",
        description="Options if configuration file does not already exist")
    init_conf_grp.add_argument('--max-concurrency', type=int,
                               help="Maximum number of concurrent jobs across "
                               "all runners")
    init_conf_grp.add_argument('--state-file-path',
                               help="File to keep Runner "
                               "state across restarts")

    parser.add_argument("-l", "--logging-level", default='INFO')

    cl_args = parser.parse_args(raw_args)
    logging.basicConfig(level=getattr(logging, cl_args.logging_level.upper()))

    conf_path = cl_args.config
    if not os.path.exists(conf_path):
        print("This is a brand new configuration file!\n")
        state_path = valid_input(
            value=cl_args.state_file_path,
            prompt="State file path",
            initial_message="Please enter state file path "
            "(used to maintain state across restarts, must be writable)")
        concurrency = valid_input(
            value=cl_args.max_concurrency,
            prompt="Concurrency",
            initial_message="Please choose maximum concurrency "
            "(across all runners to be defined in %s" % conf_path,
            converter=positive_int)
        create_conf_file(path=conf_path,
                         state_file_path=state_path,
                         concurrency=concurrency)

    coord_url = valid_input(value=cl_args.coordinator_url,
                            prompt="URL",
                            initial_message="Please enter base URL of the "
                            "GitLab / Heptapod instance to run for")
    runner_name = valid_input(value=cl_args.name,
                              prompt="Please enter runner name")
    reg_token = valid_input(value=cl_args.registration_token,
                            prompt="Please enter registration token")
    available_executors_str = "\n".join(
        "  %s (%s)" % (exe, label)
        for exe, label in AVAILABLE_EXECUTORS.items()
    )
    executor = valid_input(
        value=cl_args.executor,
        converter=valid_executor,
        prompt="Executor",
        initial_message="Please choose an executor. "
        "Available choices are:\n" + available_executors_str)

    try:
        register(conf_path=conf_path,
                 coordinator_url=coord_url,
                 runner_name=runner_name,
                 token=reg_token,
                 executor=executor)
    except RuntimeError as exc:
        logger.error("Failed to register runner: HTTP status code %d, "
                     "Body: %r", *exc.args)
        return 1
    return 0



# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later
import logging
import requests
from requests.exceptions import RequestException
import time
import toml

from .exceptions import (
    GitLabUnavailableError,
    GitLabUnexpectedError,
    JobLaunchTimeout,
)


logger = logging.getLogger(__name__)

INNER_CONFIG_EXCLUDED_ITEMS = ('heptapod_runner_main_image',
                               'paas_job_trace_watch',
                               )
"""Configuration items that are never to be forwarded to running resources.

This collection is for items that should be excluded with all concrete
subclasses of :class:`PaasRunner`.

For per-class specific rules, see :meth:`PaasRunner.is_config_item_for_inner`
"""

JOB_FAILURE_REASONS = frozenset((
    # default applied by coordinator if an unknown reason is given,
    'unknown_failure',
    'script_failure',  # ordinary job failure (no banner)
    'api_failure',
    'stuck_or_timeout_failure',
    'runner_system_failure',  # suggests in particular user to retry
    'missing_dependency_failure',
    'runner_unsupported',
    'stale_schedule',
    'job_execution_timeout',
    'archived_failure',
    'unmet_prerequisites',
    'scheduler_failure',
    'data_integrity_failure',
    'forward_deployment_failure',
    'user_blocked',
    'project_deleted',
    'ci_quota_exceeded',
    'pipeline_loop_detected',
    'trace_size_exceeded',
    'builds_disabled',
    'environment_creation_failure',
    'deployment_rejected',
    'protected_environment_failure',
    'insufficient_bridge_permissions',
    'downstream_bridge_project_not_found',
    'invalid_bridge_trigger',
    'upstream_bridge_project_not_found',
    'insufficient_upstream_permissions',
    'downstream_pipeline_creation_failed',
    'secrets_provider_not_found',
    'reached_max_descendant_pipelines_depth',
    'ip_restriction_failure',
))
"""Subset of Rails' ::Enums::Ci::CommitStatus."""


class PaasRunner:
    """Abstract base class for one of the `[[runners]]` of the main config.

    Concrete subclasses will be able to provision PAAS resources and
    launch the main Heptapod Runner's command to run one job on them.
    """

    executor = None
    """Executor name for the PAAS runner.

    Must be unique and not one of the executors provided by the main Heptapod
    Runner executable.
    """

    min_requestable_weight = 1
    """Minimum weight of job that can actually be requested from coordinator.

    If this runner is able to request the coordinator with weight limit,
    this is the weight of the smallest PAAS resource that is
    currently defined. It tells nothing about actual availability.

    On the other hand, if this runner is not able to issue requests to
    its coordinator with constraints on job weights, this must be the
    maximum weidht of all PAAS resources.

    Default value is adapted to Runners that don't have weighting
    capabilities, and excpected to be overridden by subclasses (attrivute
    or property).
    """

    @classmethod
    def register(cls):
        runner_classes[cls.executor] = cls

    @staticmethod
    def create(config):
        """Instantiate with the appropriate class."""
        return runner_classes[config['executor']](config)

    def __init__(self, config):
        self.config = config
        self.gitlab_token = self.config['token']
        # GitLab itself uses the 8 first chars in the token as a unique
        # string identifier.
        self.gitlab_api_url = self.config['url'].rstrip('/') + '/api/v4'
        self.unique_name = self.gitlab_token[:8]

    def inner_executor(self):
        """Return executor for actual use on provisoned PAAS resource.

        To be provided by concrete classes, and must be one of the executors
        provided by the main Heptapod Runner executable.

        This is a method because a given class can use several different
        executors.
        """
        raise NotImplementedError  # pragma no cover

    def is_config_item_for_inner(self, key):
        """Tell if the given item  is to be forwarded to the PAAS resource.

        For instance, configuration items allowing to provision PAAS resources
        should not be made available to a provisioned one, so that an attack
        from a job on the runner on the PAAS resource would not leak means
        to provision infinite resources.

        This method cannot be used to allow items that are in
        :const:`INNER_CONFIG_EXCLUDED_ITEMS`.

        To be extended by concrete classes
        """
        raise NotImplementedError  # pragma no cover

    def dump_inner_config(self, path):
        """Make a full configuration for use on provisioned PAAS resource.

        The produced configuration has a single `[[runner]]` section, with
        the end executor meant for the present Runner.

        It will be used with the `exec-fetched-job` command of Heptapod Runner.
        """
        # TODO should be computed and serialized (to string) only once
        inner_runner = {k: v for k, v in self.config.items()
                        if k not in INNER_CONFIG_EXCLUDED_ITEMS
                        and self.is_config_item_for_inner(k)}
        inner_runner['executor'] = self.inner_executor()
        inner_config = dict(runners=[inner_runner])
        with open(path, 'w') as fobj:
            toml.dump(inner_config, fobj)

    def request_job(self, max_weight=None):
        """Request one job from the coordinator

        :param max_weight: if the coordinator implements it, limit the
           weight of the requested job.
        :return: Job definition (JSON `str`) or `None` if there isn't any.
        """
        url = self.gitlab_api_url + '/jobs/request'
        post_data = dict(token=self.config['token'],
                         info=dict(executor=self.inner_executor(),
                                   features=dict(
                                       # always true (shells/abstract.go)
                                       upload_multiple_artifacts=True,
                                       upload_raw_artifacts=True,
                                       refspecs=True,
                                       artifacts=True,
                                       artifacts_exclude=True,
                                       multi_build_steps=True,
                                       return_exit_code=True,
                                       raw_variables=True,
                                       cache=True,
                                       masking=True,
                                       # true for Docker executor
                                       variables=True,
                                       image=True,
                                       services=True,
                                       session=True,
                                       terminal=True,
                                       # shared seems to be set to True
                                       # for shell, ssh and custom only.
                                       shared=False,
                                       # proxy seems to be set to True
                                       # for kubernetes only.
                                       # TODO evaluate if we shouldn't set
                                       # it for PAAS Runner
                                       proxy=False,
                                       ),
                                   )
                         )
        try:
            resp = requests.post(url, json=post_data)
        except RequestException as exc:
            raise GitLabUnavailableError(url=url, message=str(exc))

        if resp.status_code in (502, 503):
            # to be catched in main loop
            # (don't want to end everything if coordinator is temporarily
            # not available)
            raise GitLabUnavailableError(status_code=resp.status_code,
                                         message=resp.text,
                                         url=url)
        elif resp.status_code == 409:  # conflict
            # let's simply try later, same as a regular empty response
            logger.info("Got a 409 (Conflict) HTTP response from coordinator "
                        "while requesting a job from %s. Treating as no job",
                        url)
            return None
        elif resp.status_code >= 400:
            raise GitLabUnexpectedError(status_code=resp.status_code,
                                        params=None,
                                        message=resp.text,
                                        url=url)

        if resp.status_code == 204:  # no job
            return None
        return resp.text

    def is_job_finished(self, job_handle):
        # Using the API endpoint to get a job by its token
        url = self.gitlab_api_url + '/job'
        try:
            resp = requests.get(url, params=dict(job_token=job_handle.token))
        except RequestException as exc:
            raise GitLabUnavailableError(url=url, message=str(exc))
            # TODO raise the usual exceptions, just catch and log from
            # the polling method

        if resp.status_code == 401:
            # of course this is ugly, but assuming we are correct
            # on our call (proper token, on a job that was properly
            # acquired), then geting a 401 means that the job is not
            # running anymore. Doing something more natural may require
            # changes in the Rails app (or next main iteration of the Paas
            # Runner, with a proper HTTP service)
            return True
        elif resp.status_code >= 400:
            raise GitLabUnexpectedError(status_code=resp.status_code,
                                        params=None,
                                        message=resp.text,
                                        url=url)

        # TODO cancel provisioned resource if job is canceled
        status = resp.json().get('status')
        logger.debug("%s status is %r", job_handle, status)
        return status in ('failed', 'success', 'canceled')

    def gitlab_custom_attributes(self, resource_path, keys,
                                 token=None,
                                 resource_type='groups'):
        """Retrieve custom attributes with the given keys.

        :param token: if specified, is used instead of the usual runner token
        :returns: a `dict` with the wished keys. It is *not* guaranteed that
                  all are present.
        """
        if token is None:
            token = self.gitlab_token
        # if there are more than one keys, it is expected to be much
        # more efficient to list all attributes rather than issue as
        # many requests. This can be tweaked later if needed.
        url = '/'.join((self.gitlab_api_url,
                        resource_type,
                        resource_path.replace('/', '%2F'),
                        'custom_attributes'))
        resp = requests.get(url, headers={'Private-Token': token})
        if resp.status_code >= 400:
            raise GitLabUnexpectedError(status_code=resp.status_code,
                                        params=None,
                                        message=resp.text,
                                        url=url)
        return {attr['key']: attr['value'] for attr in resp.json()
                if attr['key'] in keys}

    def job_url(self, job_handle):
        return self.gitlab_api_url + '/jobs/%d' % job_handle.job_id

    def is_job_not_running(self, response):
        """Tell if the given API response is about a dead job."""
        return (response.status_code == 403
                and "Job is not running" in response.json().get('message', ''))

    def report_coordinator_job_failed(self, job_handle, reason):
        """Report job failure to coordinator

        :param str reason: should be one of :data:`JOB_FAILURE_REASONS`.
          If not, it will be converted by the coordinator to `unknown_failure`.
        """
        url = self.job_url(job_handle)
        if reason not in JOB_FAILURE_REASONS:
            logger.warning("Reporting job failure to coordinator with "
                           "unregistered reason %r behaves as if it were "
                           "'unknown_failure'", reason)
        params = dict(token=job_handle.token,
                      failure_reason=reason,
                      exit_code=103,
                      state='failed')
        resp = requests.put(url, json=params)

        if self.is_job_not_running(resp):
            logger.info("Reporting %s was not necessary, as it is already "
                        "seen as not running by the coordinator", job_handle)
        elif resp.status_code >= 400:
            params['token'] = 'REDACTED'
            raise GitLabUnexpectedError(status_code=resp.status_code,
                                        params=params,
                                        message=resp.text,
                                        url=url)

    def gl_job_send_trace(self, job_handle, message, content_range):
        """Inner method actually calling the coordinator.

        Separated to ease unit testing.
        """
        trace_url = self.job_url(job_handle) + '/trace'
        headers = {'content-range': content_range}
        resp = requests.patch(trace_url,
                              params=dict(token=job_handle.token),
                              headers=headers,
                              data=message
                              )
        if self.is_job_not_running(resp):
            logger.warning("Could not append trace to %s, as it is "
                           "seen as not running by the coordinator",
                           job_handle)
            return
        elif resp.status_code >= 400:
            raise GitLabUnexpectedError(status_code=resp.status_code,
                                        params=None,
                                        headers=headers,
                                        message=resp.text,
                                        url=trace_url)
        # resp.json() is an obscure integer code (seems to always be equal
        # to 3), not the new length.
        return resp.headers['Range']

    def job_append_trace(self, job_handle, message):
        """Append a message to the job trace (user-visible job log).

        :param message: the raw message, including any newline character.
        """
        length = len(message)
        new_trace_length = job_handle.trace_offset + length
        # given the checks that the coordinator performs, the returned
        # range from coordinator should be equal to ``0-{new_trace_length}``,
        # yet that is coincidental and one could imagine a later version
        # doing something else that changes the actual trace size
        # (unicode normalization or whatever), so better use what the
        # coordinator sent us.
        new_range = self.gl_job_send_trace(
            job_handle, message,
            f'{job_handle.trace_offset}-{new_trace_length}'
        )
        if new_range is not None:
            job_handle.trace_offset = int(new_range.split('-')[-1].strip())

    def job_wait_trace(self, project_id, job_handle, interruptible_sleep):
        """Wait for the job trace to have been updated outside this process.

        Can be used to check for full job startup, as typically the other
        system that would write to the job trace would be the inner runner
        on a provisioned resource.

        For now it depends on having a token able to read job traces.
        See #heptapod-runner42 for (better) future plans.

        :param interruptible_sleep: callable to sleep between calls, must
           return ``True`` if an interruption occurred (typically for
           general shutdown).
        :returns: ``True`` if a trace not written by this process
            has been detected, ``False`` if sleep between polling requests
            was interrupted.
        :raises: PaasResourceError in case of timeout
        """
        watch_conf = self.config.get('paas_job_trace_watch', {})
        token = watch_conf.get('token')
        if token is None:
            logger.warning("Missing token for job trace watching, will have "
                           "to assume it to be ok.")
            return True

        timeout = watch_conf.get('timeout_seconds', 300)
        poll_step = watch_conf.get('poll_step', 10)
        job_id = job_handle.job_id

        start = time.time()
        while time.time() - start < timeout:
            trace = self.gl_job_get_trace(project_id, job_id, token)
            if len(trace) > job_handle.trace_offset:
                logger.info("%s is fully launched (job trace update detected)",
                            job_handle)
                return True
            if interruptible_sleep(poll_step):
                return False

        raise JobLaunchTimeout(job_handle, timeout)

    def gl_job_get_trace(self, project_id, job_id, token):
        """Get the job trace for given handle, using the given token.

        In case of server side error (including 503), return an empty string.
        Because this is typically used in a polling loop, such swallowing
        is actually retrying.

        In case of 404 (Not Found) or 429 (Too Many Requests),
        an empty string is also returned for the same reasons. Note:
        according to https://docs.gitlab.com/ce/api/jobs.html#get-a-log-file,
        a 404 can be returned if there's no trace.

        :raises: :class:`GitLabUnexpectedError` in case of other client-side
          errors, as they are not supposed to happen (bad token)
        """
        trace_url = (self.gitlab_api_url
                     + f'/projects/{project_id}/jobs/{job_id}/trace')

        resp = requests.get(trace_url, headers={'Private-Token': token})

        code = resp.status_code
        if code < 400:
            return resp.text

        if code in (429, 404) or code >= 500:
            return ''

        raise GitLabUnexpectedError(url=trace_url,
                                    status_code=code,
                                    params=None,
                                    message=resp.text)

    def provision(self, job):
        """Provision necessary resources in which to actually run the job.

        To be implemented by subclasses
        """
        raise NotImplementedError('provision')  # pragma no cover

    def expected_weight(self, job):
        """Return the weight of the job without launching it.

        This is just an expected weight because it is possible (yet very
        uncommon) that the PAAS definitions change between this call
        and the actual provisioning.

        Default value is adapted to Runners that don't have weighting
        capabilities.
        """
        return 1

    def launch(self, paas_resource, job_data):
        """Schedule the job to run on the given PAAS resource.

        To be implemented by subclasses
        """
        raise NotImplementedError('launch')  # pragma no cover

    def decommission(self, paas_resource):
        """Delete the PAAS resource.

        Should not raise an error if the resource was already absent.

        :return: ``True`` if it was actually deleted, and ``False`` otherwise.
        """
        raise NotImplementedError('decommission')  # pragma no cover

    def load_paas_resource(self, data):
        """Create a full PAAS resource from minimal extracted data.

        Given a resource ``rsc`` created by this Runner instance,
        ``self.load_paas_resouce(rsc.dump())`` should give back a
        fully functional new resource.

        Typically will be used for state save before shutdown and restoration
        after startup, perhaps using changed configuration accessible from
        the Runner.
        """
        raise NotImplementedError("load_paas_resource")  # pragma no cover


runner_classes = {}
"""Mapping of executor to runner class."""



# Copyright 2021-2023 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later
import argparse
from enum import Enum
import itertools
import json
import logging
from pathlib import Path
from queue import Queue
import os
import signal
import threading
import time
import toml


# We don't need the distinction between background (bg) and foreground (fg)
from ansi.colour import fg as colour

from .exceptions import (
    ConfigurationError,
    GitLabUnavailableError,
    GitLabUnexpectedError,
    JobLaunchTimeout,
    PaasProvisioningError,
    PaasResourceError,
)
from .job import (
    JobHandle,
    jobs_log_fmt,
)
from .runner import PaasRunner

logger = logging.getLogger(__name__)


QUOTA_CONFIG_KEY = 'quota_computation'
MAX_CONCURRENCY_CONFIG_KEY = 'concurrent'


class JobEventType(Enum):
    LAUNCHING = 0
    """Job launch process has started.

    No provisioning or actual deployment have already occurred at the
    time this event is emitted.
    """

    LAUNCHED = 1
    """The job is fully running, as it can be seen from the coordinator
    """

    LAUNCH_FAILED = 2

    FINISHED = 3
    """The job is done.

    It is in particular supposed to have reported its result to coordinator
    """

    COORDINATOR_REPORT_FAILED = 4
    """Used in failure of reporting a job result to coordinator from here.

    Typically, such reporting happens only in cases of launch failures. Once
    a job is launched, it takes responsibility over the reporting to
    coordinator.
    """

    LAUNCH_REQUEST = 5
    """Used by the coordinator poll loop to request a launch.

    The corresponding message is a tuple of length 3 instead of 2.
    """

    DECOMMISSIONED = 6
    """Used to notify success of decommission for the resource of a job handle.
    """


WAKEUP_MESSAGE = object()


COORDINATOR_REPORT_LAUNCH_FAILURES_RETRY_DELAY = 30
FINAL_THREAD_SHUTDOWN_TIMEOUT = 60


SLEEP_STEP_DURATION = 2
"""
Maximum amout of time for any thread to sleep without checking for signals.

Some polling threads may sleep for a long time between their true activity.

On the
other hand, they still must check regularly for shutdown requests and the
like. Hence actual sleep time will be cut in steps of the given duration, for
signal checking wake-ups.
"""


class PaasDispatcher:
    """Main backing class for the Heptapod PAAS Runner.

    This is a multi-threaded system that polls the coordinators on
    a regular basis to acquire jobs and starts auxiliary threads to launch
    them.

    Threads:

    - :attr:`event_processing_thread`: permanent thread that listens for
      events, maintains the mutable state and starts auxiliary threads.
      Even if shutdown is required, it waits for the auxiliary threads before
      exiting. The auxiliary threads are themselves supposed to finish as
      soon as possible if shutdown is required (e.g, by breaking out of
      retry loops etc.).
    - :attr:`launched_jobs_progress_threads`: jobs (running on the PAAS
      system) report their status directoy to the coordinator. This threads
      polls the coordinator regularly to check if they are finished and
      sends appropriate events.
    - laucher threads: spawned by the event processing thread, they are
      responsible for all provisioning and startup of jobs. They report
      back with events.

    Mutable state attributes:

    - :attr:`potential_concurrency`: number of parallel jobs we'll have
      if all those the coordinator gave us get launched successfully.
      Gets compared to :attr:`max_concurrency`.

    - :attr:`potential_weight`: total weight of the jobs currently running
      or being launched (handed over by the coordinators).
      Gets compared to :attr:`weighted_quota`.

    - :attr:`shutdown_required`: boolean regularly checked by threads to
      terminate early if set. Must be set by :meth:`shutdown()` except in
      tests.

    - Job collections

      These all use :class:`JobHandle` directly or as keys.
      An important design invariant to maintain is that the event processing
      threads has write monopoly on them once it is started.

      + :attr:`pending_jobs` is a mapping from job handles to full job
        descriptions sent by the coordinator
      + :attr:`launched_jobs` contains all jobs currently fully launched
        and not yet known to be finished
      + :attr:`launch_errors`: self-explanatory

    """

    def __init__(self, config):
        self.init_runners(config)
        self.init_state_file(config)
        self.init_max_concurrency(config)
        self.init_intervals(config)
        self.finished_jobs_keep_resources = config.get(
            'paas_finished_jobs_keep_resources', False)

        self.reporting_queue = Queue()

        self.potential_weight = 0
        self.potential_concurrency = 0
        self.max_pending_jobs = config.get('paas_max_concurrent_provisioning',
                                           10)

        # job collections
        self.launch_errors = []
        self.launched_jobs = set()
        self.pending_jobs = {}
        self.to_decommission = set()

        self.launched_jobs_progress_thread = None
        self.event_processing_thread = None

        self.shutdown_required = False
        # Total number of jobs launched, successfully or not, incremented
        # when certainty about launch status has been obtained (in particular
        # does not count retries). Mostly useful for tests.
        self.total_job_launches = 0

    def init_runners(self, config):
        """Return an immutable iterable of Runner instances.

        Immutability will be helpful to avoid bugs in the requeueing loop.
        """
        # TODO catch, log then ignore init errors (perhaps in classmethod,
        # returnning None, then)
        runners = (PaasRunner.create(conf) for conf in config['runners'])
        self.runners = {runner.unique_name: runner for runner in runners}

    def init_intervals(self, config):
        self.poll_interval = config.get('check_interval', 3)
        self.job_progress_interval = config.get('job_progress_poll_interval',
                                                30)

    def runner_by_human_name(self, name):
        """Lookup runner by its human-readable name.

        The human-readable name is the `name` entry in the runner
        configuration. It is not to be confused with the runner's unique name
        (derived from its token). To lookup by `unique_name`, simply use
        the :attr:`runners` :class:dict:.

        Since unicity of the human-readable name is not formally guaranteed,
        this is a first-match logic.

        :raises KeyError: if no runner with the given human-readable name
           could be found.
        """
        for runner in self.runners.values():
            if runner.config['name'] == name:
                return runner
        raise KeyError(name)

    def init_max_concurrency(self, config):
        """Set :attr:`max_concurrency` and :attr:`weighted_quota` from config.

        Upstream GitLab Runner default value for the `concurrent` configuration
        item is 1, which doesn't make sense for a runner backed by a whole
        cloud infrastructure, so we're setting a higher, yet moderate default.
        """
        self.max_concurrency = config.get(MAX_CONCURRENCY_CONFIG_KEY, 50)
        try:
            quota = config[QUOTA_CONFIG_KEY]
        except KeyError:
            raise ConfigurationError(
                f"The {QUOTA_CONFIG_KEY} section is mandatory")

        runner_name = quota['reference_runner']
        flavor_name = quota['reference_flavor']
        njobs = quota['reference_jobs_count']
        runner = self.runner_by_human_name(runner_name)
        flavor = runner.available_flavors[flavor_name]
        self.weighted_quota = flavor.weight * njobs
        logger.warning(
            "Maximum concurrency set to %d; weighted quota set to %.1f "
            "(weight of %d jobs with flavor %s of the %r runner)",
            self.max_concurrency, self.weighted_quota,
            njobs, flavor_name, runner_name,
        )

    def init_state_file(self, config):
        path = config.get('state_file')
        if path is not None:
            path = Path(path)
        self.state_file_path = path

    def wait_all_threads(self, timeout=FINAL_THREAD_SHUTDOWN_TIMEOUT):
        logger.info("Waiting (at most %s seconds) for all "
                    "threads to finish and report back before exit",
                    timeout)

        # because the event_processing_thread waits for launcher threads
        # to finish, there's no need to wait for them directly (avoids
        # keeping a list, and related race conditions).
        for thread in (
                self.launched_jobs_progress_thread,
                self.event_processing_thread,
        ):
            if thread is not None:
                thread.join(timeout=timeout)
                if thread.is_alive():
                    logger.warning("Thread %r is still alive after "
                                   "giving it %d seconds to shut down",
                                   thread.name, timeout)

    def interruptible_sleep(self, duration, debug=''):
        """Sleep for the given duration unless there is a shutdown request.

        it will take up to :const:`SLEEP_STEP_DURATION` to detect the
        shutdown request.

        :return: whether a shutdown request was dectected
        """
        while duration > 0:
            if self.shutdown_required:
                return True
            logger.debug("interruptible sleep duration=%s %s",
                         duration, debug)
            time.sleep(min(SLEEP_STEP_DURATION, duration))
            duration -= SLEEP_STEP_DURATION
        return self.shutdown_required

    def process_events(self):
        while not self.shutdown_required:
            msg = self.reporting_queue.get()
            if msg is WAKEUP_MESSAGE:
                continue

            job_handle, status = msg[:2]
            if status is JobEventType.LAUNCH_REQUEST:
                job_data = msg[2]
                # With most of the time spent in subprocess and HTTP calls,
                # GIL expected to become a bottleneck in the long term only
                self.start_launcher_thread(job_handle, job_data)
            if status is JobEventType.LAUNCHING:
                # potential concurrency updated by the poll loop
                logger.info("Tracking the launch of %s (expected weight %.1f)",
                            job_handle, job_handle.expected_weight)
            elif status is JobEventType.LAUNCHED:
                # TODO weight correction is actually doable earlier
                # (as soon as the PAAS resource is launched, before the
                # job is fully launched. Need to come up with synonyms)
                correction = job_handle.weight_correction()
                logger.log(
                    logging.DEBUG if correction == 0 else logging.WARNING,
                    "Launched %s weight correction: %.1f",
                    job_handle, correction)
                self.potential_weight += correction
                self.launched_jobs.add(job_handle)
                self.pending_jobs.pop(job_handle, None)
                self.total_job_launches += 1
                logger.info("Successfullly launched %s (final weight %.1f)",
                            job_handle, job_handle.paas_resource.weight)
            elif status is JobEventType.LAUNCH_FAILED:
                self.potential_concurrency -= 1
                self.potential_weight -= job_handle.expected_weight
                self.pending_jobs.pop(job_handle, None)
                logger.error("Failed to launch %s", job_handle)
                self.total_job_launches += 1
                self.launch_errors.append(job_handle)
            elif status is JobEventType.COORDINATOR_REPORT_FAILED:
                # event supersedes LAUNCH_FAILED (adds info that reporting
                # the failure itself failed)
                self.potential_concurrency -= 1
                self.potential_weight -= job_handle.expected_weight
            elif status is JobEventType.FINISHED:
                logger.info("Finished %s, according to "
                            "coordinator.", job_handle)
                self.launched_jobs.discard(job_handle)
                if not self.finished_jobs_keep_resources:
                    self.to_decommission.add(job_handle)
                    self.start_decommission_thread(job_handle)
            elif status is JobEventType.DECOMMISSIONED:
                self.potential_concurrency -= 1
                self.potential_weight -= job_handle.paas_resource.weight
                self.to_decommission.discard(job_handle)
                logger.info("Resource for %s successfully decommissioned",
                            job_handle)

    def launch_job(self, job_handle, job_data):
        """Provision and start job for runner, reporting on queue

        Job arguments are very redundant to avoid needless back-and-forth
        conversions. These are, from source to most refined and partial:

        :param dict job_data: full job specification (not JSON serialized)
        :param JobHandle job_handle: representation of the job, can be
          recreated from ``job_data``.
        """

        reporting_queue = self.reporting_queue
        reporting_queue.put((job_handle, JobEventType.LAUNCHING))
        success = False

        runner = self.runners[job_handle.runner_name]
        try:
            # if job is from a previous process interrupted while the job
            # was waiting for full launch, it can have a PAAS resource, and
            # even have been launched already (with full launch not yet
            # acknowledged at the time of interruption)
            app = job_handle.paas_resource
            if app is None:
                app = runner.provision(job_data)
                job_handle.paas_resource = app
            if not app.launched:
                logger.info("Launching PAAS Ressource for %s", job_handle)
                runner.launch(app, job_data)
        except PaasProvisioningError as exc:
            logger.error("Provisioning failed for %s"
                         "(action=%r, code=%r, transport code=%r, details=%r)",
                         job_handle,
                         exc.action, exc.code, exc.transport_code,
                         exc.error_details)
            runner.job_append_trace(job_handle,
                                    colour.red(exc.error_details) + '\n')
        except PaasResourceError as exc:
            logger.error("Launching failed for %s, error on resource %r "
                         "(action=%r, code=%r, transport code=%r, details=%r)",
                         job_handle, exc.resource_id,
                         exc.action, exc.code, exc.transport_code,
                         exc.error_details)
            runner.job_append_trace(job_handle,
                                    colour.red(exc.error_details) + '\n')
        except Exception:
            logger.exception("Unexpected exception for %s", job_handle)
        else:
            try:
                if not runner.job_wait_trace(
                        job_data['job_info']['project_id'],
                        job_handle,
                        interruptible_sleep=self.interruptible_sleep
                ):
                    return
            except JobLaunchTimeout as exc:
                _, timeout = exc.args
                logger.error("No activity after launch for %s (timeout "
                             "set to %d seconds)", job_handle, timeout)
                # TODO if the job is finished (cancelled etc), then GitLab
                # will probably return a 40x error on this.
                runner.job_append_trace(
                    job_handle,
                    colour.red(f"Job not active (no log) after {timeout} "
                               f"seconds" + '\n')
                )
            else:
                success = True

        if success:
            in_process_status = JobEventType.LAUNCHED
        else:
            # script_failure is formally incorrect, but it triggers no special
            # banner rendering, whereas all other currently available values
            # (as of Heptapod 0.34.0) display a banner with
            # inappropriate messages (e.g., please retry).
            # `runner_system_failure` should be the one for
            # temporary provisioning errors, though.
            in_process_status = self.report_coordinator_job_failed(
                job_handle, reason='script_failure')
        reporting_queue.put((job_handle, in_process_status))

    def decommission(self, job_handle):
        resource = job_handle.paas_resource
        runner = self.runners[job_handle.runner_name]
        if resource is None:
            logger.error("Decommission triggered for %s, that does not "
                         "have a PAAS resource.", job_handle)
            return

        logger.info("Now decommissioning resource %r for %s",
                    resource.app_id, job_handle)
        try:
            runner.decommission(resource)
        except PaasResourceError as exc:
            logger.error("Error in decommission for %s: %r",
                         job_handle, exc)
        else:
            self.reporting_queue.put(
                (job_handle, JobEventType.DECOMMISSIONED))

    def report_coordinator_job_failed(self, job_handle, reason):
        """Report failure with exception catching and retry logic.

        :returns: the appropriate :class:JobEventType: that caller should emit.
        """
        runner = self.runners[job_handle.runner_name]
        attempt = 0
        max_attempts = 3
        delay = COORDINATOR_REPORT_LAUNCH_FAILURES_RETRY_DELAY
        while attempt < max_attempts:
            attempt += 1
            try:
                runner.report_coordinator_job_failed(job_handle, reason)
                logger.warning("Successfully reported %s as failed to launch",
                               job_handle)
                return JobEventType.LAUNCH_FAILED
            except Exception:
                logger.exception("Exception while attempting to report "
                                 "failure to launch %s to coordinator "
                                 "(attempt %d/%d). Will retry in %d seconds",
                                 job_handle, attempt, max_attempts, delay)
                if self.interruptible_sleep(delay, debug='launch_job'):
                    logger.warning("General shutdown required, stop "
                                   "retrying to launch %s", job_handle)
                    # failure to report could be due to some misconfiguration
                    # that could precisely motivation for a restart, let's
                    # keep in pending jobs so that a new process can retry
                    # the whole launching and reporting
                    break
        return JobEventType.COORDINATOR_REPORT_FAILED

    def max_pending_jobs_reached(self):
        """Tell if the maximum of pending (being launched) jobs is reached.

        This is not intended to be thread-safe, hence fully exact. It is more
        of a soft limit: it should not matter if there is one or even a couple
        of jobs acquired and not yet acknowledged by the event processing
        thread.
        """
        launching = len(self.pending_jobs)
        reached = launching >= self.max_pending_jobs
        if reached:
            logger.info("Maximum number %d of jobs being launched reached "
                        "(total %d). "
                        "Will resume polling coordinators when enough "
                        "are fully launched.",
                        self.max_pending_jobs, launching)
        return reached

    def poll_all_launch(self):
        """Poll for all runners and launch jobs.

        Each runner is polled until it doesn't get jobs to run any more.
        """
        logger.debug("poll_all_launch starting")
        event_queue = self.reporting_queue
        polling_runners = list(self.runners.values())

        while (polling_runners
               and not self.shutdown_required
               and not self.max_pending_jobs_reached()
               ):
            logger.debug("Current potential weight: %.1f, concurrency: %d",
                         self.potential_weight, self.potential_concurrency)
            next_runners = []
            for runner in polling_runners:
                headroom = self.weighted_quota - self.potential_weight
                logger.debug("Current weight headroom before polling runner: "
                             "%d (current weight of running or launching "
                             "jobs is %d, maximum configured weight is %d)",
                             headroom,
                             self.potential_weight,
                             self.max_concurrency)
                if headroom < runner.min_requestable_weight:
                    logger.info(
                        "Runner %s: not polling because the smallest weight "
                        "that can be a priori requested from coordinator is "
                        "%d, above current headroom %d (current weight "
                        "of running or launching jobs is %d, "
                        "maximum configured weight is %d)",
                        runner,
                        runner.min_requestable_weight,
                        headroom,
                        self.potential_weight,
                        self.weighted_quota)
                    continue

                if self.potential_concurrency >= self.max_concurrency:
                    continue

                logger.debug("Requesting job from runner %s",
                             runner.unique_name)
                job_json = runner.request_job(max_weight=headroom)
                if job_json is None:
                    continue

                job_data = json.loads(job_json)
                weight = runner.expected_weight(job_data)
                job_handle = JobHandle(job_id=job_data['id'],
                                       runner_name=runner.unique_name,
                                       expected_weight=weight,
                                       token=job_data['token'])
                # need to immediately update potential concurrency and weight.
                # Doing it in a thread would make it possible to acquire a
                # new job before those limits take the present one into
                # account, hence overflowing them.
                self.pending_jobs[job_handle] = job_data
                self.potential_weight += weight
                self.potential_concurrency += 1
                event_queue.put(
                    (job_handle, JobEventType.LAUNCH_REQUEST, job_data))
                # requeue for immediate repolling
                next_runners.append(runner)

            polling_runners = next_runners

    def start_launcher_thread(self, job_handle, job_data):
        launcher = threading.Thread(
            daemon=True,
            target=lambda: self.launch_job(job_handle, job_data))
        launcher.name = 'launcher-%s-%d' % (job_handle.runner_name,
                                            job_handle.job_id)
        launcher.start()

    def start_initial_threads(self):
        """Start threads for processing of initial (just loaded) state.

        Only to be used right after pending jobs have been loaded,
        and before the start of the polling loop, to avoid duplicating
        launcher threads already started by the polling loop.
        """
        # The threads we are launching can start right away mutating
        # the collections (especially in tests where waiting times can be 0).
        # Let's avoid being impaired by that:
        pending = tuple(self.pending_jobs.items())
        to_decomm = tuple(self.to_decommission)

        for job_handle, job_data in pending:
            self.start_launcher_thread(job_handle, job_data)
        for job_handle in to_decomm:
            self.start_decommission_thread(job_handle)

    def start_decommission_thread(self, job_handle):
        decom = threading.Thread(
            daemon=True,
            target=lambda: self.decommission(job_handle))
        decom.name = 'decommission-%s-%d' % (job_handle.runner_name,
                                             job_handle.job_id)
        decom.start()

    def poll_launched_jobs_progress_once(self):
        """Call coordinator to enquire about progress of launched jobs.

        This is notably useful to track down the number of currently running
        jobs.
        """
        # normally, only the reporting thread would mutate the `launched_jobs`
        # attributes, and removals are even done only upon signal from
        # this method (currently launched in a single thread).
        # Still, we can be sure of thread safety by copying to an immutable
        # structure before iteration.
        for job_handle in tuple(self.launched_jobs):
            logger.debug("Polling progress for %r", job_handle)
            runner_name = job_handle.runner_name
            runner = self.runners[runner_name]
            try:
                finished = runner.is_job_finished(job_handle)
            except GitLabUnavailableError as exc:
                # warning only because this is likely to be a temporary
                # condition
                logger.warning("Runner %r, coordinator not available, "
                               "could not poll job progress "
                               "(got %r on URL %r)",
                               runner_name, exc.message, exc.url)
            except GitLabUnexpectedError as exc:
                logger.error("Runner %r, got HTTP error %d from coordinator "
                             "while polling job progress. URL was %r, "
                             "message is %r", runner_name,
                             exc.status_code, exc.url, exc.message)
            except Exception:  # the thread must not crash
                logger.exception("Unexpected exception while polling "
                                 "coordinator for progress of %s", job_handle)
            else:
                if finished:
                    logger.warning("%r is FINISHED", job_handle)
                    self.reporting_queue.put((job_handle,
                                              JobEventType.FINISHED))

    def start_launched_jobs_progress_thread(self):
        def progress_loop():
            logger.info("Thread to poll coordinator about progress of "
                        "launched jobs started, polling every %d seconds",
                        self.poll_interval)
            while True:
                self.poll_launched_jobs_progress_once()
                if self.interruptible_sleep(self.poll_interval,
                                            debug="poll job progress"):
                    return

        thread = self.launched_jobs_progress_thread = threading.Thread(
            target=progress_loop, daemon=True)
        thread.name = "launched-jobs-progress"
        thread.start()

    def start_event_processing_thread(self):
        thread = self.event_processing_thread = threading.Thread(
            target=self.process_events, daemon=True)
        thread.name = "event-processing"
        thread.start()

    def poll_loop(self, max_cycles=None):
        """Repeatedly poll the coordinators.

        :param interval: time to wait between polling cycles if jobs limits
           are not reached
        :param interval_when_saturated: time to wait between polling cycles
           if jobs limits are reached.
        :param max_cycles: if not specified, this method never stops by
           itself. Otherwise,  the polling stops after the given
           number of cycles (note that this is not the number of time each
           coordinator gets polled.)
        """
        poll_cycles = 0
        infinite = max_cycles is None

        while infinite or poll_cycles < max_cycles:
            self.poll_all_launch()
            # No runners got job, sleep before polling coordinator again
            # TODO Slightly incorrect: we have spent some time polling for busy
            # runners, and that could become non negligible compared to the
            # wanted delay for some non-busy ones.
            # This could be fixed by another layer of threading (per
            # runner) to handle that, but that will be good enough for now.
            # Also, waiting times should be per coordinator

            if (
                    self.potential_concurrency >= self.max_concurrency
                    or self.potential_weight >= self.weighted_quota
            ):
                interval = self.job_progress_interval
                logger.info(
                    "No (more) jobs to process and max concurrency %d "
                    "or weighted quota %.1f is reached, with current %d jobs "
                    "running or being launched for a total weight of %.1f; "
                    "will poll again in %d seconds if back under limits ",
                    self.max_concurrency, self.weighted_quota,
                    self.potential_concurrency, self.potential_weight,
                    interval
                )
            else:
                interval = self.poll_interval
                logger.debug("No (more) job to process, "
                             "polling for all runners again in %d seconds",
                             interval)

            poll_cycles += 1
            if self.interruptible_sleep(interval, debug='new job loop'):
                break
        return poll_cycles

    def shutdown_signal(self, signum, frame):
        logger.warning("Caught signal %s. Triggering graceful shutdown",
                       signum)
        self.shutdown()

    def shutdown(self):
        if self.shutdown_required:
            return

        self.shutdown_required = True
        # make sure the event processing thread reconsiders the
        # shutdown flag even if there is no other remaining thread
        # to report back to it.
        self.reporting_queue.put(WAKEUP_MESSAGE)

    def log_state_signal(self, signum, frame):
        logger.warning("launched jobs=%s, pending jobs=%s,"
                       "to_decommission=%s, launch_errors=%r, "
                       "weight of acquired jobs=%.1f, weight quota=%.1f, "
                       "total acquired jobs=%d, max_concurrency=%d",
                       jobs_log_fmt(self.launched_jobs),
                       jobs_log_fmt(self.pending_jobs),
                       jobs_log_fmt(self.to_decommission),
                       self.launch_errors,
                       self.potential_weight,
                       self.weighted_quota,
                       self.potential_concurrency,
                       self.max_concurrency,
                       )

    def load_job_handle(self, data):
        runner = self.runners.get(data['runner_name'])
        return JobHandle.load(runner, data)

    def load_state(self):
        path = self.state_file_path
        if not path.exists():
            logger.info('State file "%s" does not exist. Nothing to load.',
                        path)
            return

        logger.info("Loading state from '%s'", path)
        with open(path) as fobj:
            state = json.load(fobj)

        self.launched_jobs = set(self.load_job_handle(job)
                                 for job in state['launched'])
        self.pending_jobs = {self.load_job_handle(job): data
                             for job, data in state['pending']
                             }
        self.to_decommission = set(
            self.load_job_handle(job)
            for job in state.get('to_decommission', ()))
        self.potential_concurrency = (len(state['launched'])
                                      + len(state['pending']))
        self.potential_weight = sum(
            jh.actual_or_expected_weight()
            for jh in itertools.chain(self.launched_jobs, self.pending_jobs)
        )

        logger.info("Initialized state from '%s'. Currently tracking %d "
                    "running jobs and having %d pending jobs to launch and "
                    "%d resources to decommission",
                    path, len(self.launched_jobs),
                    len(self.pending_jobs),
                    len(self.to_decommission),
                    )
        logger.debug("Full set of running jobs as loaded from '%s': %s",
                     path, jobs_log_fmt(self.launched_jobs))
        logger.info("Full set of pending jobs as loaded from '%s': %s",
                    path, jobs_log_fmt(self.pending_jobs))
        os.unlink(path)

    def save_state(self):
        path = self.state_file_path
        logger.info("Saving state to '%s'", path)
        state = dict(launched=[jh.dump() for jh in self.launched_jobs],
                     to_decommission=[jh.dump()
                                      for jh in self.to_decommission],
                     pending=[(jh.dump(), job_data)
                              for jh, job_data in self.pending_jobs.items()])
        path.touch(mode=0o600, exist_ok=True)
        with open(path, 'w') as fobj:
            json.dump(state, fobj)
        logger.info("Saved state to '%s'", path)


def main(raw_args=None):
    """Console script entry point.
    """
    parser = argparse.ArgumentParser(
        description="Second prototype for the PAAS runner system"
    )
    parser.add_argument("runner_config", help="Path to Heptapod Runner "
                        "configuration file")
    parser.add_argument("--poll-interval", type=int,
                        help="DEPRECATED Time (seconds) to wait after all "
                        "available jobs "
                        "are treated before polling coordinators again.")
    parser.add_argument("--job-progress-poll-interval", type=int,
                        help="DEPRECATED Time (seconds) to wait between "
                        "coordinators "
                        "polls about progress of successfully launched jobs")
    parser.add_argument("--poll-cycles", type=int,
                        help="Number of times to poll all runners. "
                        "(useful for testing purposes)")
    parser.add_argument("--debug-signal", action='store_true',
                        help="Dump details about current state in logs "
                        "on SIGUSR1 (can contain secrets, for debugging "
                        "purposes only)")

    parser.add_argument("-l", "--logging-level", default='INFO')

    cl_args = parser.parse_args(raw_args)
    logging.basicConfig(
        level=getattr(logging, cl_args.logging_level.upper()),
        format="%(asctime)s [%(process)d] %(name)s %(levelname)s %(message)s",
    )

    with open(cl_args.runner_config) as conf_file:
        conf = toml.load(conf_file)

    if cl_args.poll_interval is not None:
        logger.warning(
            "The --poll-interval command-line option is deprecated. "
            "Please use the `check_interval` global configuration "
            "item instead."
        )
        conf['check_interval'] = cl_args.poll_interval

    job_progress_interval = cl_args.job_progress_poll_interval
    if job_progress_interval is not None:
        logger.warning(
            "The --jobs-progress-poll-interval command-line option "
            "is deprecated. "
            "Please use the `job_progress_poll_interval` global "
            "configuration item instead."
        )
        conf['job_progress_poll_interval'] = job_progress_interval

    try:
        dispatcher = PaasDispatcher(conf)
    except ConfigurationError as exc:
        logger.fatal(exc.args[0])
        return 2

    for signum in [signal.SIGINT,
                   signal.SIGTERM,
                   ]:
        signal.signal(signum, dispatcher.shutdown_signal)
    if cl_args.debug_signal:
        signal.signal(signal.SIGUSR1, dispatcher.log_state_signal)

    dispatcher.load_state()
    dispatcher.start_initial_threads()
    dispatcher.start_event_processing_thread()
    dispatcher.start_launched_jobs_progress_thread()

    try:
        done_cycles = dispatcher.poll_loop(max_cycles=cl_args.poll_cycles)
    except Exception:
        logger.exception("Uncatched exception in main thread. Will exit "
                         "right away with abnormal termination code")
        exit_code = 1
    else:
        logger.warning("Main thread will exit normally "
                       "after %d polling cycles",
                       done_cycles)
        exit_code = 0

    dispatcher.shutdown()
    dispatcher.wait_all_threads()
    logger.debug("Done waiting for permanent threads.")
    dispatcher.save_state()
    return exit_code



# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later
import json
import logging
import pathlib
import re
import subprocess

from .exceptions import DeploymentBranchAlreadyExisting
from .version import version_str

from .job import (
    all_job_variables,
    expand_variables,
)

logger = logging.getLogger(__file__)

DOCKERFILE = (
    "FROM {base_image}\n"
    "COPY runner.toml /etc/gitlab-runner/runner.toml\n"
    "COPY job.json /etc/gitlab-runner/job.json\n"
    'ENV DEBUG="true"\n'
    'ENTRYPOINT ["/usr/bin/gitlab-runner", "exec-fetched-job", '
    '"--config", "/etc/gitlab-runner/runner.toml",'
    '"/etc/gitlab-runner/job.json"]\n'
)

DOCKER_HUB_DOMAIN = 'docker.io'
LEGACY_DOCKER_HUB_DOMAIN = 'index.docker.io'
DOCKER_HUB_OFFICIAL_NAMESPACE = 'library'
DOT_COLON_RX = re.compile(r'[.:]')


def split_docker_image_domain(image_name):
    """Split the name of a Docker image into domain and remainder.

    The image name may include a tag part or not.

    The domain part is normalized to :const:`DOCKER_HUB_DOMAIN`
    if from Docker Hub.

    Reference `Golang implementation <https://github.com/distribution/distribution/blob/main/reference/normalize.go#L87>`_::

        func splitDockerDomain(name string) (domain, remainder string) {
                i := strings.IndexRune(name, '/')
                if i == -1 || (!strings.ContainsAny(name[:i], ".:") && name[:i] != "localhost" && strings.ToLower(name[:i]) == name[:i]) {
                        domain, remainder = defaultDomain, name
                } else {
                        domain, remainder = name[:i], name[i+1:]
                }
                if domain == legacyDefaultDomain {
                        domain = defaultDomain
                }
                if domain == defaultDomain && !strings.ContainsRune(remainder, '/') {
                        remainder = officialRepoName + "/" + remainder
                }
                return
        }
    """ # noqa long lines in Golang code
    # style remark: made to follow reference impl as much as possible
    split = image_name.split('/', 1)
    remainder = split[-1]
    domain = None if len(split) == 1 else split[0]

    if domain is None or (DOT_COLON_RX.search(domain) is None
                          and not domain == 'localhost'
                          and domain.lower() == domain):
        domain, remainder = DOCKER_HUB_DOMAIN, image_name
    if domain == LEGACY_DOCKER_HUB_DOMAIN:
        domain = DOCKER_HUB_DOMAIN

    if domain == DOCKER_HUB_DOMAIN and '/' not in remainder:
        remainder = '/'.join((DOCKER_HUB_OFFICIAL_NAMESPACE,
                              remainder))
    return domain, remainder


class DockerBuildHelper:
    """Help constructing a Docker build context to run jobs for Runners.

    Attributes:

    * :attr:`path` path of the Docker build context to create
    * :attr:`git_process_env` is the common environment `dict` passed to
      `git` subprocesses. The default `None` value means to use the main
      process environment, hence in particular the global Git configuration
      of the user running the process. Passing an environment without user
      identifying variables (`USER` etc) leads to no global config being
      applied (often desirable in tests, maybe not for a service)
    * :attr:`git_user_name`, :attr:`git_user_email` are the committer name and
      email of any commit created by this builder.
      If not specified, default ones for the `git` process end up being used,
      typically inferred from environment variables and global Git
      configuration (the latter depending itself ultimately on the environment,
      see also :attr:`git_process_env`)
    """

    base_image = ("registry.heptapod.net:443/heptapod/"
                  "heptapod-runner:" + version_str)

    git_executable = 'git'

    def __init__(self, path,
                 base_image=None,
                 git_process_env=None,
                 git_user_name=None,
                 git_user_email=None):
        self.path = pathlib.Path(path)
        self.git_process_env = git_process_env
        self.git_user_name = git_user_name
        self.git_user_email = git_user_email
        if base_image is not None:
            self.base_image = base_image

    @classmethod
    def from_runner_config(cls, path, config):
        """Instantiate from a Runner configuration

        This takes care directly of common options.
        """
        return cls(path,
                   git_process_env={},
                   git_user_name="Heptapod Paas Runner",
                   git_user_email='paasrunner@heptapod.test',
                   base_image=config.get('heptapod_runner_main_image'),
                   )

    def git(self, *args, **kwargs):
        kwargs.setdefault('env', self.git_process_env)
        kwargs.setdefault('cwd', self.path)
        cmd = [self.git_executable]
        cmd.extend(args)
        return subprocess.check_output(cmd, **kwargs)

    def amend_job(self, job_data):
        self.job_use_dependency_proxy(job_data)

    def job_use_dependency_proxy(self, job_data):
        """Update job_data to use dependency proxy if available.

        The top-level group URL is used by default. If users want specific
        behaviour, they may use the other variable in the job definition.

        Reference: https://docs.gitlab.com/ee/user/packages/dependency_proxy/#authenticate-within-cicd
        """  # noqa long URL
        job_id = job_data['id']
        # TODO replace job_data by a Job class, and perform lazy
        # once-and-for-all parsing of variables.
        job_vars = all_job_variables(job_data)
        proxy_url = job_vars.get('CI_DEPENDENCY_PROXY_GROUP_IMAGE_PREFIX')
        if not proxy_url:
            logger.debug("Job %d: no dependency proxy available",
                         job_id)
            return

        main_image = job_data.get('image')
        if main_image is not None:
            main_image_name = expand_variables(main_image['name'], job_vars)
            domain, remainder = split_docker_image_domain(main_image_name)
            if domain == DOCKER_HUB_DOMAIN:
                main_image['name'] = '/'.join((proxy_url, remainder))

        services = job_data.get('services')
        if services is None:
            return

        logger.debug("Job %d: services before potential dependency "
                     "proxy rewrite %r", job_id, services)
        new_services = []
        for service in services:
            # service definition can be just a string in YaML file
            # but the coordinator normalizes it to the mapping with
            # name (image/tag), alias, entrypoint etc.
            srv_image_name = service.get('name')
            if srv_image_name is None:
                logger.warning("Job %d: did not understand image/tag "
                               "in service definition %r",
                               job_id, service)
            else:
                expanded_srv = expand_variables(srv_image_name, job_vars)
                srv_domain, srv_remainder = split_docker_image_domain(
                    expanded_srv)

                if srv_domain == DOCKER_HUB_DOMAIN:
                    service['name'] = '/'.join((proxy_url, srv_remainder))
                    # We need to insert an alias so that container linking
                    # works as originally expected by author of the job
                    # configuration (as if we didn't intercept for the
                    # dependency proxy).
                    # For this to work we need to start over from the
                    # original service name, because some normalization that
                    # would change the alias has already occurred with
                    # `srv_remainder` (`postgres` becomes `library/postgres`
                    # giving the `library-postgres` alias).

                    # At this point we are sure that if we have a host part,
                    # it is either `docker.io` or `index.docker.io`, hence
                    # without a port, so the last ':' is guaranteed to
                    # be the tag separator (also deemed 'version' in GitLab
                    # Runner sources).
                    #
                    # Reference for alias value:
                    #  https://docs.gitlab.com/ee/ci/services/#accessing-the-services
                    # We can only put one alias, so we choose the one most
                    # encouraged by GitLab docs, which is compliant with
                    # valid DNS names, aka RFC 1123 section 2.1, itself relying
                    # on RFC 952. TODO document
                    if not service.get('alias'):  # could be empty string
                        # presumably, if there is a user-defined alias, it must
                        # be in use. TODO document
                        service['alias'] = (expanded_srv.rsplit(':', 1)[0]
                                            .replace('/', '-'))
            new_services.append(service)

        job_data['services'] = new_services
        logger.debug("Job %d: services after potential dependency "
                     "proxy rewrite %r", job_id, new_services)

    def write_build_context(self, runner, job_data):
        # TODO transmit concurrency information (necessary for local
        # Docker, and more accurate job logs such as
        # "Running on runnder-TOKEN-project-ID-concurrent-NB")
        runner.dump_inner_config(self.path / 'runner.toml')
        self.amend_job(job_data)
        with (self.path / 'job.json').open('w') as jsonf:
            json.dump(job_data, jsonf)
        (self.path / 'Dockerfile').write_text(
            DOCKERFILE.format(base_image=self.base_image))

    def git_push(self, url):
        """Make the Docker context a Git repository, commit and push."""
        self.git('init', '-q')
        if self.git_user_name:
            self.git('config', 'user.name', self.git_user_name)
        if self.git_user_email:
            self.git('config', 'user.email', self.git_user_email)

        self.git('add', '.')
        self.git('commit', '-m', 'Job definition')
        remote = 'paas'
        self.git('remote', 'add', remote, url)
        # TODO expected remote branch could become, e.g, 'main'
        remote_branch = 'master'
        if self.git('ls-remote', remote, remote_branch):
            raise DeploymentBranchAlreadyExisting(url, remote_branch)
        self.git('push', '-q', '--set-upstream', remote, remote_branch)



# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later
import os
from overrides import overrides
from urllib.parse import urlparse
import subprocess
import tempfile

from .docker import DockerBuildHelper
from .exceptions import (
    PaasResourceError,
)
from .runner import PaasRunner

# same environment variable supported by the Docker executor, see
# https://docs.gitlab.com/runner/configuration/advanced-configuration.html#the-runnersdocker-section
DEFAULT_DOCKER_URL = os.environ.get('DOCKER_HOST',
                                    'unix:///var/run/docker.sock')


class LocalDockerRunner(PaasRunner):

    executor = 'local-docker'

    def __init__(self, config):
        super(LocalDockerRunner, self).__init__(config)

        docker_raw_url = self.config.get('host', DEFAULT_DOCKER_URL)
        docker_url = urlparse(docker_raw_url)
        if docker_url.scheme != 'unix':
            raise ValueError("Runner %r (executor %r) invalid Docker host "
                             "URL %r: only Unix Domain sockets "
                             "(unix:///some/path.socket) are supported " % (
                                 self.unique_name, self.executor,
                                 docker_raw_url))
        self.docker_socket = docker_url.path

    @overrides
    def inner_executor(self):
        return 'docker'

    @overrides
    def is_config_item_for_inner(self, key):
        """All config parameters except `host` are standard.

        `host`, the Docker endpoint URL is meaningful for the dispatcher only,
        and is mounted at a fixed place for the inner executor.
        """
        return key != 'host'

    @overrides
    def provision(self, job):
        app_id = 'hpd-job-%s-%d' % (self.unique_name.lower(), job['id'])
        return LocalDockerApplication(app_id)

    @overrides
    def load_paas_resource(self, data):
        return LocalDockerApplication(data)

    @overrides
    def launch(self, paas_resource, job_data):
        """Build the image and run the job, like an actual PAAS runner would.

        TODO avoid subprocess, and use the docker Python lib, so that this
        method becomes testable (if only with severe mocking, but at least
        could catch some discrepancies on our side).
        """
        docker_image = paas_resource.docker_image
        try:
            with tempfile.TemporaryDirectory() as tmp_path:
                build_helper = DockerBuildHelper.from_runner_config(
                    tmp_path, self.config)
                build_helper.write_build_context(self, job_data)
                subprocess.check_call(("docker", "build",
                                       '-t', docker_image,
                                       tmp_path))

                subprocess.check_call(
                    ("docker", "run",
                     '--rm', '-t',
                     '-v', self.docker_socket + ':/var/run/docker.sock',
                     docker_image))
        except Exception as exc:
            raise PaasResourceError(executor=self.executor,
                                    action='launch',
                                    resource_id=paas_resource.app_id,
                                    code=None,
                                    error_details=repr(exc),
                                    )

    @overrides
    def decommission(self, paas_resource):
        subprocess.check_call(('docker', 'rmi', paas_resource.docker_image))


LocalDockerRunner.register()


class LocalDockerApplication:

    def __init__(self, app_id):
        """By convention, the application is the Docker image."""
        self.app_id = self.docker_image = app_id

    def dump(self):
        return self.app_id



# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later
import logging
from overrides import overrides
import requests
import tempfile
import time

from .docker import DockerBuildHelper
from .exceptions import (
    DeploymentBranchAlreadyExisting,
    GitLabUnexpectedError,
    PaasProvisioningError,
    PaasResourceError,
)
from .job import get_job_variable
from .runner import PaasRunner

logger = logging.getLogger(__name__)

FLAVOR_JOB_VARIABLE_NAME = 'CI_CLEVER_CLOUD_FLAVOR'
DEFAULT_FLAVOR = 'M'
"""Ultimate default if flavor not specified in job nor runner config."""

CC_ORGA_ID_ATTRIBUTE = 'cc_orga_id'
CC_ORGA_TOKEN_ATTRIBUTE = 'cc_orga_token'


def convert_http_error(http_response):
    """Return error code, message."""
    try:
        resp_data = http_response.json()
        assert resp_data['type'] == 'error'
        # checked with Clever, `id` is definitely an error code
        return resp_data['id'], resp_data['message']
    except Exception:
        return None, http_response.text


class CleverCloudDockerRunner(PaasRunner):

    executor = 'clever-docker'
    _available_flavors = (None, None)
    """dict of available flavors and timestamp of retrieval.
    """

    def __init__(self, config):
        super(CleverCloudDockerRunner, self).__init__(config)
        self.init_cc_global()
        self.init_cc_git()
        self.init_cc_orga()
        self.init_common_app_params()

    def init_cc_global(self):
        self.cc_base_url = self.config.get('cc_api_url',
                                           'https://api.clever-cloud.com/v2')
        self.cc_zone = self.config.get('cc_zone', 'par')
        self.cc_default_flavor = self.config.get('cc_default_flavor',
                                                 DEFAULT_FLAVOR)
        self.cc_max_flavor = self.config.get('cc_max_flavor')
        # TODO validate early the existence of the max flavor
        # what if it disappears afterwards?

        self.deployment_repo_timeout = self.config.get(
            'cc_deployment_repo_timeout', 20)
        self.deployment_repo_wait_step = self.config.get(
            'cc_deployment_repo_wait_step', 2)

    def init_cc_orga(self):
        if self.config.get('cc_multi_tenant'):
            self.fixed_cc_orga = None
        else:
            self.fixed_cc_orga = CleverCloudOrganization(
                self.cc_base_url,
                orga_id=self.config['cc_orga_id'],
                git_user=self.cc_git_user,
                token=self.config['cc_token'])
        self.gl_attributes_token = self.config.get(
            'cc_gitlab_namespace_attributes_token')
        self.cc_orga_id_attribute = self.config.get(
            'cc_orga_id_attribute', CC_ORGA_ID_ATTRIBUTE)
        self.cc_orga_token_attribute = self.config.get(
            'cc_orga_token_attribute', CC_ORGA_TOKEN_ATTRIBUTE)
        self.all_cc_orga_attributes = (self.cc_orga_id_attribute,
                                       self.cc_orga_token_attribute,
                                       )

    @property
    def available_flavors(self):
        """Cached dict of available flavors.

        This is a mapping whose keys are user flavor names, such as `'XL'`
        and values are :class:`CleverCloudFlavor` instances

        The caching is not really thread-safe, but the worst that can
        happen is several GET requests towards the Clever API if several
        threads realize that the timestamp is outdated at the same time.
        It is very unlikely that these requests would give different results
        and even more unlikely that it would make a difference for the end
        user (using a flavor that just got advertised or that doesn't exist
        anymore. In the first case, the job will be able to run earlier than
        expected, in the second case we're only losing a job that would have
        been invalid mere seconds later).
        """
        available, ts = self._available_flavors
        if available is not None and time.time() - ts < 3600:
            return available

        timestamp = time.time()
        # default values should be the final ones in production,
        # configurability will help us in particular to prototype with
        # the Jenkins flavors.
        flavor_context = self.config.get('cc_flavor_context',
                                         'heptapod-runner')
        params = dict(context=flavor_context)
        resp = requests.get(self.cc_base_url + '/products/flavors',
                            params=params)

        if resp.status_code >= 400:
            # TODO surely this must most of the times
            # be a transient failure. Retry logic?
            error_code, message = convert_http_error(resp)
            raise PaasProvisioningError(action='get-available-flavors',
                                        action_details=params,
                                        transport_code=resp.status_code,
                                        code=error_code,
                                        error_details=message,
                                        executor=self.executor)

        prefix = self.config.get('cc_flavor_name_prefix', flavor_context + '-')
        available = {
            flav['name'][len(prefix):]: CleverCloudFlavor(flav)
            for flav in resp.json()
            if flav['name'].startswith(prefix) and flav['available']}
        logger.info("Retrieved variant flavors. Resulting mapping: %r",
                    available)
        self._available_flavors = available, timestamp
        return available

    @property
    def min_requestable_weight(self):
        if self.cc_max_flavor is not None:
            return self.available_flavors[self.cc_max_flavor].weight

        return max(fl.weight for fl in self.available_flavors.values())

    def init_cc_git(self):
        self.cc_git_user = self.config.get('cc_auth_user', 'Jenkins')

    def init_common_app_params(self):
        self.common_app_params = {
            # TODO could depend on job rather than runner
            'zone': self.cc_zone,
            'instanceLifetime': 'TASK',
            'instanceType': 'docker',
            'deploy': 'git',
            'instanceVariant': self.config.get(
                'cc_instance_variant_uuid',
                # This is the initial UUID for Heptapod Runner variant
                "70d676a0-a775-4dc9-a43d-3121036e8515"),
            'maxInstances': 1,  # we'll do our own auto scaling
            'minInstances': 1,
        }

    def inner_executor(self):
        return 'docker'

    def is_config_item_for_inner(self, key):
        return (key != 'heptapod_runner_main_image'
                and not key.startswith('cc_'))

    def cc_docker_instance_type(self):
        """Retrieve the CC instance type that we'll be using, with details.

        The version is crucial, as we need to specify a supported one upon
        provisioning. See also #37 about caching.
        """
        resp = requests.get(self.cc_base_url + '/products/instances')
        if resp.status_code >= 400:
            error_code, message = convert_http_error(resp)
            raise PaasProvisioningError(action='get-instance-type',
                                        transport_code=resp.status_code,
                                        code=error_code,
                                        error_details=message,
                                        executor=self.executor)

        assert resp.status_code == 200
        docker_types = [inst for inst in resp.json()
                        if inst['name'] == 'Docker']
        # TODO what if several match? Are, e.g., several versions
        # possible at a given time?
        return docker_types[0]

    def cc_select_flavor(self, instance_type, job, check_max=True):
        flavor = get_job_variable(job, FLAVOR_JOB_VARIABLE_NAME)
        if flavor is None:
            flavor = self.cc_default_flavor

        available = self.available_flavors
        cc_flavor = available.get(flavor)
        if cc_flavor is None:
            raise PaasProvisioningError(
                action='check-flavor',
                transport_code=None,
                code=1,
                error_details="Selected flavor %r is not available, "
                "possible choices are %r" % (flavor,
                                             list(self.available_flavors)),
                executor=self.executor,
            )
        if check_max and self.cc_max_flavor is not None:
            max_flavor = self.available_flavors[self.cc_max_flavor]
            if cc_flavor.weight > max_flavor.weight:
                raise PaasProvisioningError(
                    action='check-flavor',
                    transport_code=None,
                    code=2,
                    error_details=f"Flavor size exceeded: this runner is "
                    f"configured to have a maximum Clever Cloud flavor of "
                    f"'{self.cc_max_flavor}', but this job specifies "
                    f"the bigger flavor '{flavor}'. Please change the value "
                    f"of the {FLAVOR_JOB_VARIABLE_NAME} job variable or "
                    f"select another runner if possible.",
                    executor=self.executor,
                )

        return cc_flavor

    def job_cc_orga(self, job):
        orga = self.fixed_cc_orga
        if orga is not None:
            return orga

        return self.cc_orga(self.gitlab_job_top_namespace(job))

    def gitlab_job_top_namespace(self, job):
        return get_job_variable(job, 'CI_PROJECT_ROOT_NAMESPACE')

    def cc_orga(self, gitlab_namespace):
        orga = self.fixed_cc_orga
        if orga is not None:
            return orga

        # perhaps we should have a dedicated PaasAccessError
        # meanwhile, this will be good enough.
        try:
            attrs = self.gitlab_custom_attributes(
                gitlab_namespace,
                self.all_cc_orga_attributes,
                token=self.gl_attributes_token,
            )
        except GitLabUnexpectedError as exc:
            raise PaasProvisioningError(
                executor=self.executor,
                action='find-orga',
                action_details='namespace=%r, '
                'failed attributes request on %r: %s' % (
                    gitlab_namespace,
                    exc.url,
                    exc.message
                ),
                code=exc.status_code)
        try:
            orga_id, orga_token = (attrs[self.cc_orga_id_attribute],
                                   attrs[self.cc_orga_token_attribute])
        except KeyError as exc:
            raise PaasProvisioningError(
                executor=self.executor,
                action='find-orga',
                action_details='namespace=%r, '
                'missing attribute %r' % (gitlab_namespace,
                                          exc.args[0]),
                code=None)

        return CleverCloudOrganization(
            self.cc_base_url,
            gitlab_namespace=gitlab_namespace,
            orga_id=orga_id,
            git_user=self.cc_git_user,
            token=orga_token)

    @overrides
    def provision(self, job):
        # TODO cache instance_type, and reload only if provision ends
        # with errors possibly due to outdated information (including change
        # of flavor availability)
        # (error code for unknown application should be 4004)
        instance_type = self.cc_docker_instance_type()

        req_data = self.common_app_params.copy()
        req_data['instanceVersion'] = instance_type['version']

        flavor = self.cc_select_flavor(instance_type, job)
        req_data['minFlavor'] = req_data['maxFlavor'] = flavor.api_name

        req_data['name'] = 'hpd-job-%s-%d' % (self.unique_name, job['id'])
        app = self.job_cc_orga(job).create_app(req_data)

        cc_env = {'CC_MOUNT_DOCKER_SOCKET': 'true'}
        extra_env = self.config.get('cc_extra_env')
        if extra_env:
            cc_env.update(extra_env)

        app.put_env(cc_env)
        return app

    @overrides
    def expected_weight(self, job):
        # TODO find a way to cache this call (with proper invalidation):
        instance_type = self.cc_docker_instance_type()
        # we'll let the launching thread check the max flavor and
        # provide end-user feedback.
        flavor = self.cc_select_flavor(instance_type, job, check_max=False)
        return flavor.weight

    @overrides
    def load_paas_resource(self, data):
        # this maintains compatibility with data dumped without a
        # GitLab namespace for runners with a fixed CC Organization
        gl_namespace = data.get('gitlab_namespace')
        try:
            cc_orga = self.cc_orga(gl_namespace)
        except PaasProvisioningError as exc:
            # In a multi-tenant system, one misconfigured tenant
            # should not crash the whole system.
            logger.error(
                "Failed to load PAAS resource from data %r."
                "Operations on resource, notably decommission, "
                "won't be possible. "
                "action=%r, action_details=%r, code=%r, error_details=%r",
                data, exc.action, exc.action_details,
                exc.code, exc.error_details)
            # we should also not lose data about the resource just
            # because we couldn't resolve the orga this time
            cc_orga = OrganizationNotFound(gl_namespace)

        return cc_orga.load_application(data)

    def wait_deployability(self, app):
        app_id = app.app_id
        wait_step = self.deployment_repo_wait_step
        timeout = self.deployment_repo_timeout

        start = time.time()
        while (time.time() - start) < timeout:
            try:
                state = app.get_info()['deployment']['repoState']
            except KeyError:
                raise PaasResourceError(
                    app_id, self.executor,
                    action='wait-deployability',
                    error_details=f"Clever Cloud Application {app_id}: "
                    "deployment details schema not understood",
                    code=3,
                    transport_code=None,
                )
            if state == 'CREATED':
                break
            time.sleep(wait_step)
        else:
            raise PaasResourceError(
                app_id, self.executor,
                action='wait-git',
                error_details='Clever Cloud Application '
                f'{app_id}: not deployable after {timeout} seconds',
                code=4,
                transport_code=None
            )

    @overrides
    def launch(self, paas_resource, job_data):
        self.wait_deployability(paas_resource)

        with tempfile.TemporaryDirectory() as tmp_path:
            build_helper = DockerBuildHelper.from_runner_config(tmp_path,
                                                                self.config)
            build_helper.write_build_context(self, job_data)
            try:
                build_helper.git_push(paas_resource.git_push_url)
            except DeploymentBranchAlreadyExisting:
                logger.warning(
                    "%s: depoyment branch already existing on remote. "
                    "There probably was an issue with state loading "
                    "or shutdown (SIGKILL or bug in shutdown sequence)",
                    paas_resource.log_fmt()
                )

        paas_resource.launched = True

    @overrides
    def decommission(self, paas_resource):
        app_id = paas_resource.app_id

        resp = paas_resource.orga.delete_app(app_id)
        if resp.status_code == 404:
            logger.warning("decommision: application %r is already "
                           "not present, nothing to be done", app_id)
            return False

        if resp.status_code >= 400:
            error_code, message = convert_http_error(resp)
            raise PaasResourceError(app_id, self.executor,
                                    action='delete',
                                    action_details=None,
                                    code=error_code,
                                    transport_code=resp.status_code,
                                    error_details=message)
        logger.debug('decommission: successful response %r', resp.text)
        return True

    def decommission_all(self, gitlab_namespace=None):
        """Decommission all PAAS resources for this runner.

        This is obviously dangerous and if ever exposed to the CLI or
        users in any way, care must be taken to make sure it can't
        happen by accident.

        :param gitlab_namespace: if specified, will be used to
           find the relevant Organization. Otherwise, :attr:`fixed_cc_orga`
           will be used. If the latter is ``None`` (typically case of a
           multi-tenant Runner), then `ValueError` is raised.

        :returns: the number of decommissionned resources and number of
           ignored ones (not related to this runner).
        """
        if gitlab_namespace is None:
            if self.fixed_cc_orga is None:
                raise ValueError("This Runner %r does not have a fixed "
                                 "Clever Cloud Organization. "
                                 "To decommission all "
                                 "its resources, the path of the GitLab "
                                 "namespace bearing the CC Organization and "
                                 "token must be given." % self.unique_name)
            else:
                cc_orga = self.fixed_cc_orga
        else:
            cc_orga = self.cc_orga(gitlab_namespace)

        logger.warning("Decommissionning all PAAS resources for runner %r"
                       " on Clever Cloud Organization %r",
                       self.unique_name, cc_orga.orga_id)
        rsc_name_prefix = 'hpd-job-%s-' % self.unique_name
        done = 0
        ignored = 0
        for app_data in cc_orga.list_applications():
            if not app_data.get('name', '').startswith(rsc_name_prefix):
                ignored += 1
                continue

            app = cc_orga.application(app_data)
            logger.info("Decommissioning app name=%r", app_data['name'])
            self.decommission(app)
            time.sleep(10)
            done += 1

        logger.info("Decommissionned the %d applications spawned by "
                    "runner %r, and ignored %d unrelated applications "
                    "in the same organization %r",
                    done, self.unique_name, ignored, cc_orga.orga_id)
        return done, ignored


CleverCloudDockerRunner.register()


class CleverCloudOrganization:

    executor = CleverCloudDockerRunner.executor
    """Used in logging and raising of exceptions."""

    transitional_default_weight = 32768
    """Weight to attribute to applications created before weight tracking.

    This is a worst-case value, meant to avoid over-provisionning due
    to underestimation of currently running jobs at the first restart
    with the weighted quota system.

    We could do better than that, by querying actual flavors in service
    upon state load, but it isn't worth the effort at this point: all we'll
    have if there are lots of running jobs is a temporary saturation, until
    at least one finishes.
    """

    def __init__(self, base_api_url, orga_id, token, git_user,
                 gitlab_namespace=None):
        """Information needed to work on Clever API for an Organization.
        """
        self.base_api_url = base_api_url
        self.orga_id = orga_id
        self.git_user = git_user
        self.token = token
        self.gitlab_namespace = gitlab_namespace

        self.url = '/'.join((base_api_url, 'organisations', orga_id))
        self.headers = {
            'Content-Type': 'application/json',
            'Authorization': 'token ' + self.token,
        }

    @property
    def applications_url(self):
        return '/'.join((self.url, 'applications'))

    def app_url(self, app_id):
        return '/'.join((self.applications_url, app_id))

    def create_app(self, data):
        """Create an application by calling the proper endpoint.

        :returns: :class:`CleverCloudApplication`
        """
        logger.debug("create_app data=%r", data)
        resp = requests.post(self.applications_url,
                             headers=self.headers,
                             json=data)
        if resp.status_code >= 400:
            # TODO catch error due to outdated version to recompute it
            # and raise something appropriate for other errors.
            # Actual error got due to lack of version is currently in use in
            # test_clever.py for easy reference.
            error_code, message = convert_http_error(resp)
            raise PaasProvisioningError(action='create-app',
                                        action_details=data,
                                        code=error_code,
                                        transport_code=resp.status_code,
                                        error_details=message,
                                        executor=self.executor,
                                        )
        logger.debug("CleverCloud application creation: "
                     "successful response %r", resp.text)

        return self.application(resp.json())

    def application(self, resp_data):
        """Represent an existing CC application from API response data.

        The data can be obtained, e.g., from a creation request, or a listing
        request.
        """
        return CleverCloudApplication(
            orga=self,
            app_id=resp_data['id'],
            deploy_url=resp_data['deployment']['httpUrl'],
            user=self.git_user,
            password=self.token,
            weight=CleverCloudFlavor(resp_data['instance']['maxFlavor']).weight
        )

    def load_application(self, data):
        """Load application from saved data."""
        return CleverCloudApplication(
            orga=self,
            app_id=data['id'],
            deploy_url=data['deploy_url'],
            user=self.git_user,
            launched=data['launched'],
            password=self.token,
            weight=data.get('weight', self.transitional_default_weight),
        )

    def delete_app(self, app_id):
        """Delete an application.

        Kept as a method on the organization, because we need less
        detailed information to delete an application than to use it.
        """
        return requests.delete(self.app_url(app_id),
                               headers=self.headers)

    def list_applications(self):
        """Return the list of all applications.

        This list may include applications that are *not* related to
        a given Runner, or not even CI related: this is the Clever Cloud
        API request result after JSON decoding.
        """
        return requests.get(self.applications_url,
                            headers=self.headers).json()


class OrganizationNotFound(CleverCloudOrganization):
    """Represent a failure to resolve an organization.

    Used to keep track of CleverCloudApplication instances even if
    the corresponding organization could not be resolved (typically after
    a restart on a wrong configuration or while the coordinator is down).
    This is enough to allow various kinds of retry.
    """

    def __init__(self, gitlab_namespace):
        self.gitlab_namespace = gitlab_namespace
        self.git_user = None
        self.token = None
        self.url = None

    def __eq__(self, other):
        return self.gitlab_namespace == other.gitlab_namespace

    def app_url(self, app_id):
        """Always return None to express URL is not known."""

    def delete_app(self, app_id):
        raise PaasResourceError(
            app_id, self.executor,
            action='delete',
            code=22,  # chosen arbitrarily (randint call)
            error_details="Could not find Clever Cloud Organization "
            "credentials for the GitLab namespace %r this resource is "
            "linked to. Earlier logs may have details "
            "about failed attempts." % self.gitlab_namespace,
            )


class CleverCloudApplication:

    def __init__(self, orga, app_id, deploy_url, user, password,
                 launched=False,
                 weight=None):
        self.orga = orga
        self.app_id = app_id
        self.api_url = orga.app_url(self.app_id)
        self.deploy_url = deploy_url

        self.git_push_url = deploy_url.replace(
            'https://',
            'https://%s:%s@' % (user, password))
        self.weight = weight
        self.launched = launched

    @property
    def executor(self):
        return self.orga.executor

    @property
    def headers(self):
        return self.orga.headers

    def dump(self):
        # we don't dump credentials. If this is loaded by a Runner,
        # it will reinitialize the credentials, maybe with an updated value.
        return dict(id=self.app_id,
                    gitlab_namespace=self.orga.gitlab_namespace,
                    weight=self.weight,
                    launched=self.launched,
                    deploy_url=self.deploy_url)

    def log_fmt(self):
        """String representation useful for debug logs.

        Not included attributes or fields:

        - :attr:`deploy_url`: can be found by requesting the CC API if needed.
        - :attr:`gitlab_namespace`: can be found by looking at the job in the
          GitLab UI (or API)
        """
        return (f'CleverCloudApplication(app_id={self.app_id!r}, '
                f'weight={self.weight:.2f}, '
                f'launched={self.launched})')

    def get_info(self):
        resp = requests.get(self.api_url,
                            headers=self.headers)
        if resp.status_code >= 400:
            error_code, message = convert_http_error(resp)
            raise PaasResourceError(self.app_id, self.executor,
                                    action='put-env',
                                    action_details=None,
                                    code=error_code,
                                    transport_code=resp.status_code,
                                    error_details=message)
        return resp.json()

    def put_env(self, env):
        """Add environment variables to the application.
        """
        resp = requests.put(self.api_url + '/env',
                            json=env,
                            headers=self.headers)

        if resp.status_code >= 400:
            error_code, message = convert_http_error(resp)
            raise PaasResourceError(self.app_id, self.executor,
                                    action='put-env',
                                    action_details=env,
                                    code=error_code,
                                    transport_code=resp.status_code,
                                    error_details=message)
        # success code should be 328.
        logger.debug('CleverCloudApplication.put_env: successful response %r',
                     resp.text)
        return resp


class CleverCloudFlavor:
    def __init__(self, api_data):
        """Initialize from data returned by Clever Cloud API.

        Could be from the flavor listing request, or from one of the flavors
        listed in an application

        Attributes:

        - ``api_name``: the actual names expected by Clever Cloud API
          (e.g, `'heptapod-runner-XL'`)
        - ``ram``: allocated RAM, in MiB
        - ``cpus``: number of virtual cores
        """
        # TODO actual response also includes a detailed `memory` key with
        # values such as {'unit': 'B', value: '4294967296'}
        # using it could be more robust than relying on the fact that `mem`
        # is in MiB (we'd just need to implement all decimal and binary units
        # (e.g., GB, kiB etc)
        self.api_name = api_data['name']
        self.ram_mib = api_data['mem']
        self.cpus = api_data['cpus']

    @property
    def weight(self):
        """Based on RAM footprint, with bias favoring small instances.

        The reason for the bias is that it is typically harder to fit
        large virtual machines in the available room on hypervisors. Of course
        this merely helping, not providing any kind of certainty about the
        provisionability.

        The current formula makes it so that a flavor taking 16 times the
        RAM of a smaller one counts 2 times as much as 16 of the smaller.
        """
        return (self.ram_mib / 1024) ** 1.25

    def __repr__(self):
        return "CleverCloudFlavor({'name': '%s', 'mem': %d, 'cpus': %d})" % (
            self.api_name, self.ram_mib, self.cpus
        )



from pathlib import Path
from setuptools import setup

MAIN_PACKAGE = 'heptapod_runner'

scripts = {  # name to (module in main package, function)
    'heptapod-paas-runner': ('paas_dispatcher', 'main'),
    'heptapod-paas-runner-register': ('paas_register', 'main'),
}

VERSION_FILE = 'VERSION'

with open('install-requirements.txt', 'r') as install_reqf:
    install_req = [req.strip() for req in install_reqf]


setup(
    name='python-heptapod-runner',
    version=Path(MAIN_PACKAGE, VERSION_FILE).read_text().strip(),
    author='Georges Racinet',
    author_email='georges.racinet@octobus.net',
    url='https://foss.heptapod.net/heptapod/runner',
    description="Heptapod Runner: Python utilities and subsystems",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    keywords='hg mercurial git heptapod gitlab',
    license='GPLv3+',
    # do not use find_packages, as it could recurse into the Git and
    # Mercurial repositories
    packages=[MAIN_PACKAGE],
    package_data={MAIN_PACKAGE: [VERSION_FILE]},
    entry_points=dict(
        console_scripts=[
            '{name}={pkg}.{mod}:{fun}'.format(
                pkg=MAIN_PACKAGE, name=name, mod=mod, fun=fun)
            for name, (mod, fun) in scripts.items()],
    ),
    install_requires=install_req
)



# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later
import io
import json
from overrides import overrides
import requests

from .runner import PaasRunner


COORDINATOR_URL = 'https://gitlab.test'
COORDINATOR_JOB_REQUEST_URL = 'https://gitlab.test/api/v4/jobs/request'


def request_recorder(records, responses):
    """Record sent requests and treat each from the given responses.

    :param responses: a list of :class:`request.Response` instances or
                      :class:`Exception` instances. In the latter case,
                      the exception is raised instead of returning the
                      response.
    :param records: list to which the recorder appends the request that is
                    being sent.
    """
    def request(meth, url, **kwargs):
        records.append((meth, url,  kwargs))
        resp = responses.pop(0)
        if isinstance(resp, Exception):
            raise resp
        return resp

    return request


def make_response(status_code, body,
                  headers=None,
                  encoding='utf-8'):
    """Create a :class:`request.Response` instance.

    :param body: must be `bytes`.
    """
    resp = requests.models.Response()
    resp.status_code = status_code
    # force encoding as suggested in docstring of requests.models
    # because init from headers doesn't work with the way we instantiate
    # the response, leading to a detection library being used, that fires
    # a UserWarning if content is too small (facepalm unicode religion)
    resp.encoding = encoding
    resp.raw = io.BytesIO(body)
    if headers is not None:
        resp.headers = headers
    return resp


def make_json_response(obj, status_code=200, headers=None):
    return make_response(status_code, json.dumps(obj).encode(),
                         headers=headers)


class PaasResource():
    def __init__(self, rsc_id, credentials):
        self.id = rsc_id
        self.credentials = credentials

    def dump(self):
        """Like real-workd PAAS resources, we don't include credentials."""
        return self.id


class RunnerForTests(PaasRunner):

    paas_credentials = None

    def inner_executor(self):
        return 'docker'

    def is_config_item_for_inner(self, key):
        return not key.startswith('priv_')

    def paas_resource(self, rsc_id):
        return PaasResource(rsc_id, self.paas_credentials)

    @overrides
    def load_paas_resource(self, data):
        return PaasResource(data, self.paas_credentials)


RunnerForTests.register()



# Copyright 2021 Georges Racinet <georges.racinet@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-3.0-or-later
from io import StringIO
import re


class JobHandle:
    """Represent a job provided by a coordinator.

    - :attr:`job_id` the numeric id, relative to the coordinator
    - :attr:`runner_name` the unique name of the Runner instance that
      was used to acquire the job
    - :attr:`token` authentication token to the coordinator.

    This class is meant to be message passing payload, hence does not
    have heavy attributes.

    With the Runner instance (usually can be retrieved from its unique name)
    and :attr:`token`, all jobs details can be retrieved from coordinator.
    Some of them (status, notably) can change over time.
    We could therefore use ``(runner_name, token)`` as a full ID, but
    :attr:`token` must stay secret and never be leaked to the logs, which
    could otherwise happen easily on changes of this class.
    """

    paas_resource = None
    """Attribute to keep track of provisioned resource, if any."""

    def __init__(self, runner_name, job_id, token,
                 trace_offset=0,
                 expected_weight=1):
        self.runner_name = runner_name
        self.job_id = job_id
        self.token = token
        self.full_id = (runner_name, job_id)
        self.expected_weight = expected_weight
        self.trace_offset = trace_offset

    @classmethod
    def load(cls, runner, data):
        rsc_data = data.pop('paas_resource', None)
        handle = cls(**data)
        if rsc_data is not None:
            handle.paas_resource = runner.load_paas_resource(rsc_data)
        return handle

    def weight_correction(self):
        return self.paas_resource.weight - self.expected_weight

    def actual_or_expected_weight(self):
        if self.paas_resource is not None:
            return self.paas_resource.weight
        return self.expected_weight

    def dump(self):
        res = {attr: getattr(self, attr)
               for attr in ('runner_name',
                            'job_id',
                            'token',
                            'trace_offset',
                            'expected_weight')}
        if self.paas_resource is not None:
            res['paas_resource'] = self.paas_resource.dump()
        return res

    def __eq__(self, other):
        return (isinstance(other, self.__class__)
                and self.full_id == other.full_id)

    def __hash__(self):
        return hash(self.full_id)

    def __repr__(self):
        return 'JobHandle(runner_name=%r, job_id=%d, token="REDACTED")' % (
            self.runner_name, self.job_id)

    def __str__(self):
        return 'job %d for runner %r' % (self.job_id, self.runner_name)

    def log_fmt(self):
        """Alternative to str() and repr() useful to log pending jobs."""
        rsc = self.paas_resource
        rsc_log = None if rsc is None else rsc.log_fmt()
        return (f'JobHandle<id={self.job_id}, runner={self.runner_name!r}, '
                f'resource={rsc_log}>')


def jobs_log_fmt(collection):
    """Format for logs a collection of jobs.

    :param collectio: any iterable of job handles. Notably formats the
      *keys* of :class:`dict`s such as ``PaasDispatcher.pending_jobs``.
    """
    stream = StringIO()
    stream.write('[')
    first = True
    for handle in collection:
        if not first:
            stream.write(', ')
        stream.write(handle.log_fmt())
        first = False
    stream.write(']')
    return stream.getvalue()


def get_job_variable(job_data, name):
    """Extract the variable with given name from job data dict.

    :returns: the variable value as string or ``None`` if it doesn't exist.
    """
    for var in job_data.get('variables', ()):
        if var['key'] == name:
            return var['value']


def all_job_variables(job_data):
    """Return a `dict for all variables in job_data.

    All values stay simple strings.
    """
    return {v['key']: v['value'] for v in job_data.get('variables', ())}


VAR_SUBST_RX = re.compile(r'\$([a-zA-Z0-9_]+|{.+?})')
r"""Regular expression for variables substitution

Implementation notes:

- not using `\w` to avoid a doc lookup for people wondering about edge cases
  (underscore, dash) and to avoid the need for `re.ASCII` option. See also
  https://docs.python.org/3/howto/regex.html#matching-characters
"""


def expand_variables(s, job_vars):
    """Expand given string with the given job variables.

    This is meant to be the same as GitLab Runner would do, with
    substitution syntax being either:

    - ``$`` followed by alphanumerics (stop at first non-alphanum)
    - or ``${`` up to the first ``}`` (ignoring nesting, clearly)

    (this is what Gos' ``os.ExpandVar`` does

    See heptapod-runner#25 for more details.

    Pretty much everything is understood with the braces notation::

      >>> expand_variables('somethin-${FOO-BAR}-hop', {'FOO-BAR': "foo"})
      'somethin-foo-hop'
      >>> expand_variables('somethin-${FOO-BAR}-hop', {'FOO-BAR': "foo"})
      'somethin-foo-hop'
      >>> expand_variables('somethin-${F#*$&}-hop', {'F#*$&': "foo"})
      'somethin-foo-hop'

    Without braces, variable names in pattern are restricted to
    alphanumeric ASCII chars and underscores::

      >>> expand_variables('somethin-$FOO_BAR-hop', {'FOO_BAR': "foo"})
      'somethin-foo-hop'
      >>> expand_variables('$FOO_BAR-hop', dict(FOO_BAR="foo"))
      'foo-hop'
      >>> expand_variables('somethin-$FOO-BAR-hop', {'FOO-BAR': "foo"})
      'somethin--BAR-hop'

    Special cases for uniformity with Golang implementation (checked with
    the online playfield at https://pkg.go.dev/os#Expand)::

      >>> expand_variables('somethin-${FOO-${BAR}}-hop', {'FOO-BAR': "foo"})
      'somethin-}-hop'
      >>> expand_variables('somethin-${FOO-${BAR}}-hop', {'FOO-${BAR': "foo"})
      'somethin-foo}-hop'

    Logically nested (but not syntactically)::

      >>> expand_variables('somethin-${FOO}-hop',
      ...                  {'FOO': "foo${BAR}", 'BAR': "bar"})
      'somethin-foobar-hop'
    """
    def repl(match):
        # perhaps a more clever regexp would prevent the need to strip braces
        return expand_variables(job_vars.get(match.group(1).strip('{}'), ''),
                                job_vars)

    return VAR_SUBST_RX.sub(repl, s)
